//------------------------------------------------------------------------------------------------
//
//   Greg's Mod Base - Generic Mod
//
//------------------------------------------------------------------------------------------------

package gcewing.prospecting;

import java.io.*;
import java.lang.reflect.*;
import java.net.*;
import java.util.*;
import java.util.jar.*;

import net.minecraft.block.*;
import net.minecraft.creativetab.*;
import net.minecraft.entity.player.*;
import net.minecraft.inventory.*;
import net.minecraft.item.*;
import net.minecraft.util.*;
import net.minecraft.world.*;

import net.minecraftforge.common.*;
import net.minecraftforge.client.*;
import net.minecraftforge.oredict.*;

import cpw.mods.fml.common.*;
import cpw.mods.fml.common.event.*;
import cpw.mods.fml.common.network.*;
import cpw.mods.fml.common.registry.*;
import cpw.mods.fml.relauncher.ReflectionHelper;

public class BaseMod implements IGuiHandler {

	public BaseConfiguration config;
	public String modPackage;
	public String resourceDir; // path to resources directory with leading and trailing slashes
	public String textureFile; // path to default texture file with leading slash
	public URL resourceURL; // URL to the resources directory
	public BaseMod base;
	public IGuiHandler proxy;
	public boolean serverSide, clientSide;
	//public static String channel;

	static File cfgFile;

	public String resourcePath(String fileName) {
		return resourceDir + fileName;
	}

	public BaseMod(String pkg) {
		base = this;
		modPackage = pkg;
		String resourceRelDir = modPackage.replace(".", "/") + "/resources/";
		resourceDir = "/" + resourceRelDir;
		textureFile = resourceDir + "textures.png";
		resourceURL = getClass().getClassLoader().getResource(resourceRelDir);
		//System.out.printf("BaseMod: resourceURL = %s\n", resourceURL);
	}

	//@Mod.PreInit
	public void preInit(FMLPreInitializationEvent e) {
		serverSide = e.getSide().isServer();
		clientSide = e.getSide().isClient();
		cfgFile = e.getSuggestedConfigurationFile();
		loadConfig();
		boolean[] configMarkers = ReflectionHelper.getPrivateValue(Configuration.class, config, "configMarkers");
		preallocateBlockIDs(configMarkers);
		preallocateItemIDs(configMarkers);
	}
	
	void preallocateBlockIDs(boolean[] configMarkers) {
		ConfigCategory items = config.getCategory(config.CATEGORY_BLOCK);
		for (Property prop : items.getValues().values()) {
			int id = prop.getInt();
			if (id != -1) {
				System.out.printf("BaseMod.preallocateItemIDs: Marking block id %d as in use\n", id);
				configMarkers[id] = true;
			}
		}
	}
	
	void preallocateItemIDs(boolean[] configMarkers) {
		ConfigCategory items = config.getCategory(config.CATEGORY_ITEM);
		for (Property prop : items.getValues().values()) {
			int id = prop.getInt();
			if (id != -1) {
				System.out.printf("BaseMod.preallocateItemIDs: Marking item id %d as in use\n", id);
				configMarkers[id + 256] = true;
			}
		}
	}
	
	//@Mod.Init
	public void init(FMLInitializationEvent e) {
		MinecraftForge.EVENT_BUS.register(this);
	}
	
	//@Mod.PostInit
	public void postInit(FMLPostInitializationEvent e) {
		registerBlocks();
		registerItems();
		registerOres();
		registerRecipes();
		registerTileEntities();
		registerRandomItems();
		registerWorldGenerators();
		if (e.getSide().isClient())
			proxy = initClient();
		else
			proxy = this;
		NetworkRegistry.instance().registerGuiHandler(this, proxy);
		saveConfig();
	}

	void loadConfig() {
		config = new BaseConfiguration(cfgFile);
	}

	void saveConfig() {
		if (config.extended)
			config.save();
	}

	String qualifiedName(String name) {
		return modPackage + "." + name;
	}
	
	//--------------- Proxy creation ----------------------------------------------------------

	BaseModClient initClient() {
		return new BaseModClient(this);
	}

	//--------------- Item construction ----------------------------------------------------------
	
	Item newItem(String name, String title) {
		return newItem(name, Item.class, title);
	}

	<ITEM extends Item> ITEM newItem(String name, Class<ITEM> cls, String title) {
		try {
			int id = config.getItem(name, 31743).getInt();
//			Constructor<ITEM> ctor = cls.getConstructor(int.class, String.class);
//			ITEM item = ctor.newInstance(id, modPackage + ":" + name);
//			item.setUnlocalizedName(qualifiedName(name));
			Constructor<ITEM> ctor = cls.getConstructor(int.class);
			ITEM item = ctor.newInstance(id);
			item.setUnlocalizedName(modPackage + ":" + name);
			LanguageRegistry.addName(item, title);
			if (clientSide) {
				if (item.getCreativeTab() == null)
					item.setCreativeTab(CreativeTabs.tabMisc);
			}
			System.out.printf("BaseMod.newItem: %s unlocalizedName = %s title = %s\n",
				item, item.getUnlocalizedName(), title);
			return item;
		}
		catch (Exception e) {
			throw new RuntimeException(e);
		}
	}
	
	//--------------- Block construction ----------------------------------------------------------

	Block newBlock(String name, String title) {
		return newBlock(name, Block.class, title);
	}
	
	<BLOCK extends Block> BLOCK newBlock(String name, Class<BLOCK> cls, String title) {
		return newBlock(name, cls, ItemBlock.class, title);
	}
	
	<BLOCK extends Block> BLOCK newBlock(String name, Class<BLOCK> cls, Class itemClass, String title) {
		try {
			int id = config.getBlock(name, 4095).getInt();
			Constructor<BLOCK> ctor = cls.getConstructor(int.class);
			BLOCK block = ctor.newInstance(id);
			//if (block instanceof BaseBlock) {
			//	((BaseBlock)block).modName = modPackage;
			//	((BaseBlock)block).textureName = name;
			//}
			block.setUnlocalizedName(modPackage + ":" + name);
			GameRegistry.registerBlock(block, itemClass);
			LanguageRegistry.addName(block, title);
			if (block.getCreativeTabToDisplayOn() == null)
				block.setCreativeTab(CreativeTabs.tabMisc);
			return block;
		}
		catch (Exception e) {
			throw new RuntimeException(e);
		}
	}
	
	//--------------- Ore registration ----------------------------------------------------------

	void registerOre(String name, Block block) {
		OreDictionary.registerOre(name, new ItemStack(block));
	}
	
	void registerOre(String name, Item item) {
		OreDictionary.registerOre(name, new ItemStack(item));
	}

	//--------------- Recipe construction ----------------------------------------------------------

	void newRecipe(Item product, int qty, Object... params) {
		newRecipe(new ItemStack(product, qty), params);
	}
	
	void newRecipe(Block product, int qty, Object... params) {
		newRecipe(new ItemStack(product, qty), params);
	}

	void newRecipe(ItemStack product, Object... params) {
		GameRegistry.addRecipe(new ShapedOreRecipe(product, params));
	}

	void newShapelessRecipe(Block product, int qty, Object... params) {
		newShapelessRecipe(new ItemStack(product, qty), params);
	}
	
	void newShapelessRecipe(Item product, int qty, Object... params) {
		newShapelessRecipe(new ItemStack(product, qty), params);
	}
	
	void newShapelessRecipe(ItemStack product, Object... params) {
		GameRegistry.addRecipe(new ShapelessOreRecipe(product, params));
	}

	void newSmeltingRecipe(Item product, int qty, Item input) {
		GameRegistry.addSmelting(input.itemID, new ItemStack(product, qty), 0);
	}
	
	void newSmeltingRecipe(Item product, int qty, Block input) {
		GameRegistry.addSmelting(input.blockID, new ItemStack(product, qty), 0);
	}
	
	//--------------- Dungeon loot ----------------------------------------------------------

	void addRandomChestItem(ItemStack stack, int minQty, int maxQty, int weight, String... category) {
		WeightedRandomChestContent item = new WeightedRandomChestContent(stack, minQty, maxQty, weight);
		for (int i = 0; i < category.length; i++)
			ChestGenHooks.addItem(category[i], item);
	}

	//--------------- Method stubs ----------------------------------------------------------

	void registerBlocks() {}
	void registerItems() {}
	void registerOres() {}
	void registerRecipes() {}
	void registerTileEntities() {}
	void registerRandomItems() {}
	void registerWorldGenerators() {}
	
	//--------------- Debugging ----------------------------------------------------------

	public Set<String> listResources(String subdir) {
		try {
			Set<String>result = new HashSet<String>();
			String protocol = resourceURL.getProtocol();
			if (protocol.equals("jar")) {
				String resPath = resourceURL.getPath();
				int pling = resPath.indexOf("!");
				URL jarURL = new URL(resPath.substring(0, pling));
				String resDirInJar = resPath.substring(pling + 2);
				String prefix = resDirInJar + subdir + "/";
				//System.out.printf("BaseMod.listResources: looking for names starting with %s\n", prefix);
				JarFile jar = new JarFile(new File(jarURL.toURI()));
				Enumeration<JarEntry> entries = jar.entries();
				while (entries.hasMoreElements()) {
					String name = entries.nextElement().getName();
					if (name.startsWith(prefix) && !name.endsWith("/") && !name.contains("/.")) {
						//System.out.printf("BaseMod.listResources: name = %s\n", name);
						result.add(name.substring(prefix.length()));
					}
				}
			}
			else
				throw new RuntimeException("Resource URL protocol " + protocol + " not supported");
			return result;
		}
		catch (Exception e) {
			throw new RuntimeException(e);
		}
	}
	
	//--------------- GUIs ----------------------------------------------------------

	void openGui(EntityPlayer player, Enum id, World world, int x, int y, int z) {
		openGui(player, id.ordinal(), world, x, y, z);
	}

	void openGui(EntityPlayer player, int id, World world, int x, int y, int z) {
		player.openGui(base, id, world, x, y, z);
	}
	
	Map<Integer, Class<? extends Container>> containerClasses =
		new HashMap<Integer, Class<? extends Container>>();

	void registerContainers() {
	}
	
	public void addContainer(Enum id, Class<? extends Container> cls) {
		addContainer(id.ordinal(), cls);
	}

	public void addContainer(int id, Class<? extends Container> cls) {
		containerClasses.put(id, cls);
	}
	
	/**
	 * Returns a Container to be displayed to the user. 
	 * On the client side, this needs to return a instance of GuiScreen
	 * On the server side, this needs to return a instance of Container
	 *
	 * @param ID The Gui ID Number
	 * @param player The player viewing the Gui
	 * @param world The current world
	 * @param x X Position
	 * @param y Y Position
	 * @param z Z Position
	 * @return A GuiScreen/Container to be displayed to the user, null if none.
	 */

	@Override
	public Object getServerGuiElement(int id, EntityPlayer player, World world, int x, int y, int z) {
		Class cls = containerClasses.get(id);
		if (cls != null)
			return createGuiElement(cls, player, world, x, y, z);
		else
			return getGuiContainer(id, player, world, x, y, z);
	}
	
	@Override
	public Object getClientGuiElement(int id, EntityPlayer player, World world, int x, int y, int z) {
		return null;
	}
	
	Object createGuiElement(Class cls, EntityPlayer player, World world, int x, int y, int z) {
		try {
			try {
				System.out.printf("BaseMod.createGuiElement: Invoking create method of %s for %s in %s\n",
					cls, player, world);
				return cls.getMethod("create", EntityPlayer.class, World.class, int.class, int.class, int.class)
					.invoke(null, player, world, x, y, z);
			}
			catch (NoSuchMethodException e) {
				System.out.printf("BaseMod.createGuiElement: Invoking constructor of %s\n", cls);
				return cls.getConstructor(EntityPlayer.class, World.class, int.class, int.class, int.class)
					.newInstance(player, world, x, y, z);
			}
		}
		catch (Exception e) {
			Throwable cause = e.getCause();
			System.out.printf("BaseMod.createGuiElement: %s: %s\n", e, cause);
			if (cause != null)
				cause.printStackTrace();
			else
				e.printStackTrace();
			//throw new RuntimeException(e);
			return null;
		}
	}
	
	Container getGuiContainer(int id, EntityPlayer player, World world, int x, int y, int z) {
		//  Called when container id not found in registry
		return null;
	}

}
